# coding: utf-8
#

import hashlib
import logging
import os
import shutil
import tarfile

import humanize
import progress.bar
import requests
from logzero import logger, setup_logger
from retry import retry
from uiautomator2.version import __apk_version__, __atx_agent_version__, __jar_version__

appdir = os.path.join(os.path.expanduser("~"), '.uiautomator2')

GITHUB_BASEURL = "https://github.com/openatx"


class DownloadBar(progress.bar.PixelBar):
    message = "Downloading"
    suffix = '%(current_size)s/%(total_size)s'
    width = 10

    @property
    def total_size(self):
        return humanize.naturalsize(self.max, gnu=True)

    @property
    def current_size(self):
        return humanize.naturalsize(self.index, gnu=True)


def cache_download(url, filename=None, timeout=None, logger=logger):
    """ return downloaded filepath """
    # check cache
    if not filename:
        filename = os.path.basename(url)
    storepath = os.path.join(appdir,
                             hashlib.sha224(url.encode()).hexdigest(),
                             filename)
    storedir = os.path.dirname(storepath)
    if not os.path.isdir(storedir):
        os.makedirs(storedir)
    if os.path.exists(storepath) and os.path.getsize(storepath) > 0:
        logger.debug("Use cached assets: %s", storepath)
        return storepath

    logger.debug("Download %s", url)
    # download from url
    headers = {
        'Accept': '*/*',
        'Accept-Encoding': 'gzip, deflate, br',
        'Accept-Language': 'zh-CN,zh;q=0.8,zh-TW;q=0.7,zh-HK;q=0.5,en-US;q=0.3,en;q=0.2',
        'Connection': 'keep-alive',
        'Origin': 'https://github.com',
        'User-agent': 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_14_3) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/73.0.3683.86 Safari/537.36'
    } # yapf: disable
    r = requests.get(url, stream=True, headers=headers, timeout=None)
    r.raise_for_status()

    file_size = int(r.headers.get("Content-Length"))
    bar = DownloadBar(filename, max=file_size)
    with open(storepath + '.part', 'wb') as f:
        chunk_length = 16 * 1024
        while 1:
            buf = r.raw.read(chunk_length)
            if not buf:
                break
            f.write(buf)
            bar.next(len(buf))
        bar.finish()

    assert file_size == os.path.getsize(storepath + ".part")
    shutil.move(storepath + '.part', storepath)
    return storepath


def mirror_download(url, filename: str, logger=logger):
    github_host = "https://github.com"
    if url.startswith(github_host):
        mirror_url = "https://tool.appetizer.io" + url[len(
            github_host):]  # mirror of github
        try:
            return cache_download(mirror_url,
                                  filename,
                                  timeout=60,
                                  logger=logger)
        except requests.RequestException as e:
            logger.debug("download mirror err: %s, use origin source", e)

    return cache_download(url, filename, logger=logger)


class Initer():
    def __init__(self, device, loglevel=logging.INFO):
        d = self._device = device

        self.sdk = d.getprop('ro.build.version.sdk')
        self.abi = d.getprop('ro.product.cpu.abi')
        self.pre = d.getprop('ro.build.version.preview_sdk')
        self.arch = d.getprop('ro.arch')
        self.abis = (d.getprop('ro.product.cpu.abilist').strip()
                     or self.abi).split(",")
        self.server_addr = None
        self.logger = setup_logger(level=loglevel)
        self.logger.debug("Initial device %s", device)

    def shell(self, *args):
        self.logger.debug("Shell: %s", args)
        return self._device.shell(args)

    @property
    def apk_urls(self):
        """
        Returns:
            iter([name, url], [name, url])
        """
        for name in ["app-uiautomator.apk", "app-uiautomator-test.apk"]:
            yield (name, "".join([
                GITHUB_BASEURL,
                "/android-uiautomator-server/releases/download/",
                __apk_version__, "/", name
            ]))
    
    @property
    def jar_urls(self):
        """
        Returns:
            iter([name, url], [name, url])
        """
        for name in ['bundle.jar', 'uiautomator-stub.jar']:
            yield (name, "".join([
                GITHUB_BASEURL,
                "/android-uiautomator-jsonrpcserver/releases/download/",
                __jar_version__, "/", name
            ]))

    @property
    def atx_agent_url(self):
        files = {
            'armeabi-v7a': 'atx-agent_{v}_linux_armv7.tar.gz',
            'arm64-v8a': 'atx-agent_{v}_linux_armv7.tar.gz',
            'armeabi': 'atx-agent_{v}_linux_armv6.tar.gz',
            'x86': 'atx-agent_{v}_linux_386.tar.gz',
        }
        name = None
        for abi in self.abis:
            name = files.get(abi)
            if name:
                break
        if not name:
            raise Exception(
                "arch(%s) need to be supported yet, please report an issue in github"
                % self.abis)
        return GITHUB_BASEURL + '/atx-agent/releases/download/%s/%s' % (
            __atx_agent_version__, name.format(v=__atx_agent_version__))

    @property
    def minicap_urls(self):
        base_url = GITHUB_BASEURL + \
            "/stf-binaries/raw/master/node_modules/minicap-prebuilt/prebuilt/"
        sdk = self.sdk
        yield base_url + self.abi + "/lib/android-" + sdk + "/minicap.so"
        yield base_url + self.abi + "/bin/minicap"

    @property
    def minitouch_url(self):
        return ''.join([
            GITHUB_BASEURL + "/stf-binaries",
            "/raw/master/node_modules/minitouch-prebuilt/prebuilt/",
            self.abi + "/bin/minitouch"
        ])

    def push_url(self, url, dest=None, mode=0o755, tgz=False, extract_name=None):  # yapf: disable
        path = mirror_download(url, os.path.basename(url), logger=self.logger)
        if tgz:
            tar = tarfile.open(path, 'r:gz')
            path = os.path.join(os.path.dirname(path), extract_name)
            tar.extract(extract_name, os.path.dirname(path))
        if not dest:
            dest = "/data/local/tmp/" + os.path.basename(path)

        self.logger.debug("Push to %s:0%o", dest, mode)
        self._device.sync.push(path, dest, mode=mode)
        return dest

    def is_apk_outdated(self):
        apk1 = self._device.package_info("com.github.uiautomator")
        if not apk1:
            return True
        if apk1['version_name'] != __apk_version__:
            return True
        if not self._device.package_info("com.github.uiautomator.test"):
            return True
        return False

    def is_atx_agent_outdated(self):
        agent_version = self._device.shell("/data/local/tmp/atx-agent version").strip()
        if agent_version == "dev":
            self.logger.info("skip version check for atx-agent dev")
            return False

        # semver major.minor.patch
        try:
            real_ver = list(map(int, agent_version.split(".")))
            want_ver = list(map(int, __atx_agent_version__.split(".")))
        except ValueError:
            return True

        self.logger.debug("Real version: %s, Expect version: %s", real_ver,
                          want_ver)

        if real_ver[:2] != want_ver[:2]:
            return True

        return real_ver[2] < want_ver[2]

    def check_install(self):
        """
        Only check atx-agent and test apks (Do not check minicap and minitouch)

        Returns:
            True if everything is fine, else False
        """
        d = self._device
        if d.sync.stat("/data/local/tmp/atx-agent").size == 0:
            return False

        if self.is_atx_agent_outdated():
            return False

        packages = d.list_packages()
        if 'com.github.uiautomator' not in packages:
            return False
        if 'com.github.uiautomator.test' not in packages:
            return False

        return True

    def _install_apks(self):
        """ use uiautomator 2.0 to run uiautomator test """
        self.shell("pm", "uninstall", "com.github.uiautomator")
        self.shell("pm", "uninstall", "com.github.uiautomator.test")
        for _, url in self.apk_urls:
            path = self.push_url(url, mode=0o644)
            package_name = "com.github.uiautomator.test" if "test.apk" in url else "com.github.uiautomator"
            if os.getenv("TMQ"):
                # used inside TMQ platform
                self.shell(
                    "CLASSPATH=/sdcard/tmq.jar", "exec", "app_process",
                    "/system/bin",
                    "com.android.commands.monkey.other.InstallCommand",
                    "-r", "-v", "-p", package_name, path)
            else:
                self.shell("pm", "install", "-r", "-t", path)
    
    def _install_jars(self):
        """ use uiautomator 1.0 to run uiautomator test """
        for (name, url) in self.jar_urls:
            self.push_url(url, "/data/local/tmp/"+name, mode=0o644)
    
    def _install_atx_agent(self, server_addr=None):
        self.logger.info("Install atx-agent %s", __atx_agent_version__)
        self.push_url(self.atx_agent_url,
                            tgz=True,
                            extract_name="atx-agent")
        args = ["/data/local/tmp/atx-agent", "server", "--nouia", "-d"]
        if server_addr:
            args.extend(['-t', server_addr])
        self.shell("/data/local/tmp/atx-agent", "server", "--stop")
        self.shell(*args)

    def install(self, server_addr=None):
        self.logger.info("Install minicap, minitouch")
        self.push_url(self.minitouch_url)
        if self.abi == "x86":
            self.logger.info(
                "abi:x86 seems to be android emulator, skip install minicap")
        elif int(self.sdk) >= 29:
            self.logger.info("Android Q (sdk:29) has no minicap resource")
        else:
            for url in self.minicap_urls:
                self.push_url(url)

        self._install_jars()
        if self.is_apk_outdated():
            self.logger.info(
                "Install com.github.uiautomator, com.github.uiautomator.test %s",
                __apk_version__)
            self._install_apks()
        else:
            self.logger.info("Already installed com.github.uiautomator apks")

        if self.is_atx_agent_outdated():
            self._install_atx_agent(server_addr)

        self.logger.info("Check install")
        self.check_atx_agent_version()
        print("Successfully init %s" % self._device)

    @retry(
        (requests.ConnectionError, requests.ReadTimeout, requests.HTTPError),
        delay=.5,
        tries=10)
    def check_atx_agent_version(self):
        port = self._device.forward_port(7912)
        self.logger.debug("Forward: local:tcp:%d -> remote:tcp:%d", port, 7912)
        version = requests.get("http://127.0.0.1:%d/version" % port).text.strip()
        self.logger.debug("atx-agent version %s", version)

    def uninstall(self):
        self._device.shell(["/data/local/tmp/atx-agent", "server", "--stop"])
        self._device.shell(["rm", "/data/local/tmp/atx-agent"])
        self._device.shell(["rm", "/data/local/tmp/minicap"])
        self._device.shell(["rm", "/data/local/tmp/minicap.so"])
        self._device.shell(["rm", "/data/local/tmp/minitouch"])
        self._device.shell(["am", "uninstall", "com.github.uiautomator"])
        self._device.shell(["am", "uninstall", "com.github.uiautomator.test"])


if __name__ == "__main__":
    import adbutils

    serial = None
    device = adbutils.adb.device(serial)
    init = Initer(device, loglevel=logging.DEBUG)
    print(init.check_install())



# coding: utf-8
#

from uiautomator2 import utils

def test_list2cmdline():
    testdata = [
        [("echo", "hello"), "echo hello"],
        [("echo", "hello&world"), "echo 'hello&world'"],
        [("What's", "your", "name?"), """'What'"'"'s' your 'name?'"""]
    ]
    for args, expect in testdata:
        cmdline = utils.list2cmdline(args)
        assert cmdline == expect, "Args: %s, Expect: %s, Got: %s" % (args, expect, cmdline)


import json
import os
import datetime
import atexit

from uiautomator2 import UIAutomatorServer
from uiautomator2.ext.info import conf


class Info(object):
    def __init__(self, driver, package_name=None):
        self._driver = driver
        self.output_dir = 'report/'
        self.pkg_name = package_name
        self.test_info = {}
        atexit.register(self.write_info)

    def read_file(self, filename):
        try:
            with open(self.output_dir + filename, 'r') as f:
                return f.read()
        except IOError as e:
            print(os.strerror(e.errno))

    def get_basic_info(self):
        device_info = self._driver.device_info
        app_info = self._driver.app_info(self.pkg_name)
        # query for exact model info
        if device_info['model'] in conf.phones:
            device_info['model'] = conf.phones[device_info['model']]
        self.test_info['basic_info'] = {'device_info': device_info, 'app_info': app_info}

    def get_app_icon(self):
        icon = self._driver.app_icon(self.pkg_name)
        icon.save(self.output_dir + 'icon.png')

    def get_record_info(self):
        record = json.loads(self.read_file('record.json'))
        steps = len(record['steps'])
        start_time = datetime.datetime.strptime(record['steps'][0]['time'],
                                                '%H:%M:%S')
        end_time = datetime.datetime.strptime(
            record['steps'][steps - 1]['time'], '%H:%M:%S')
        total_time = end_time - start_time
        self.test_info['record_info'] = {
            'steps': steps,
            'start_time': record['steps'][0]['time'],
            'total_time': str(total_time)
        }

    def get_result_info(self):
        log = self.read_file('log.txt')
        trace_list = []
        if log:
            log = log.splitlines()
            for i in range(len(log)):
                if 'Traceback' in log[i]:
                    new_trace = log[i]
                    i += 1
                    while 'File' in log[i]:
                        new_trace += '\n' + log[i]
                        i += 1
                    new_trace += '\n' + log[i]
                    trace_list.append(new_trace)
        self.test_info['trace_info'] = {
            'trace_count': len(trace_list),
            'trace_list': trace_list
        }

    def start(self):
        self.get_basic_info()
        self.get_app_icon()

    def write_info(self):
        # self.get_basic_info()
        self.get_record_info()
        self.get_result_info()
        with open(self.output_dir + 'info.json', 'wb') as f:
            f.write(json.dumps(self.test_info))



# coding: utf-8
#
# Refs:
# - https://opencv-python-tutroals.readthedocs.io/en/latest/

import base64
import logging
import os
import re
import time
from typing import Union

import cv2
import findit
import numpy as np
import requests
from logzero import setup_logger
from PIL import Image


def pil2cv(pil_image):
    """ Convert from pillow image to opencv """
    # convert PIL to OpenCV
    pil_image = pil_image.convert('RGB')
    cv2_image = np.array(pil_image)
    # Convert RGB to BGR
    cv2_image = cv2_image[:, :, ::-1].copy()
    return cv2_image


def cv2pil(cv_image):
    """ Convert opencv to pillow image """
    return Image.fromarray(cv_image[:, :, ::-1].copy())


def _open_data_url(data, flag=cv2.IMREAD_COLOR):
    pos = data.find('base64,')
    if pos == -1:
        raise IOError("data url is invalid, head %s" % data[:20])

    pos += len('base64,')
    raw_data = base64.decodestring(data[pos:])
    image = np.asarray(bytearray(raw_data), dtype="uint8")
    image = cv2.imdecode(image, flag)
    return image


def _open_image_url(url: str, flag=cv2.IMREAD_COLOR):
    """ download the image, convert it to a NumPy array, and then read
    it into OpenCV format """
    content = requests.get(url).content
    image = np.asarray(bytearray(content), dtype="uint8")
    image = cv2.imdecode(image, flag)
    return image


def imread(data):
    """
    Args:
        data: local path or http url or data:image/base64,xxx
    
    Returns:
        opencv image
    
    Raises:
        IOError
    """
    if isinstance(data, np.ndarray):
        return data
    elif isinstance(data, Image.Image):
        return pil2cv(data)
    elif data.startswith('data:image/'):
        return _open_data_url(data)
    elif re.match(r'^https?://', data):
        return _open_image_url(data)
    elif os.path.isfile(data):
        im = cv2.imread(data)
        if im is None:
            raise IOError("Image format error: %s" % data)
        return im

    raise IOError("image read invalid data: %s" % data)


class ImageX(object):
    def __init__(self, d: "uiautomator2.Device"):
        """
        Args:
            d (uiautomator2 instance)
        """
        self.logger = setup_logger()
        self._d = d
        assert hasattr(d, 'click')
        assert hasattr(d, 'screenshot')

        # self.logger.setLevel(logging.INFO)

    def send_click(self, x, y):
        return self._d.click(x, y)

    def match(self, imdata: Union[np.ndarray, str]):
        """
        Args:
            imdata: file, url, pillow or opencv image object
        
        Returns:
            templateMatch result
        """
        cvimage = imread(imdata)
        fi = findit.FindIt(engine=['template'],
                           engine_template_scale=(0.9, 1.1, 3),
                           pro_mode=True)
        fi.load_template("template", pic_object=cvimage)

        target = self._d.screenshot(format='opencv')
        raw_result = fi.find("target", target_pic_object=target)
        result = raw_result['data']['template']['TemplateEngine']
        target_sim = result['target_sim']  # 相似度  similarity

        return {"similarity": target_sim, "point": result['target_point']}

    def __wait(self, imdata, timeout=30.0, threshold=0.8):
        deadline = time.time() + timeout
        while time.time() < deadline:
            m = self.match(imdata)
            sim = m['similarity']
            self.logger.debug("similarity %.2f [~%.2f], left time: %.1fs", sim,
                              threshold, deadline - time.time())
            if sim < threshold:
                continue
            time.sleep(.1)
            return m

    def wait(self, imdata, timeout=30.0):
        m = self.__wait(imdata, timeout=timeout, threshold=0.8)
        if m is None:
            return m
        # time.sleep(.1)
        return self.__wait(imdata, timeout=timeout, threshold=0.9)

    def click(self, imdata, timeout=30.0):
        """
        Args:
            imdata: file, url, pillow or opencv image object
        """
        res = self.wait(imdata, timeout=timeout)
        if res is None:
            raise RuntimeError("image object not found")
        x, y = res['point']
        return self.send_click(x, y)


if __name__ == "__main__":
    im = imread("https://www.baidu.com/img/bd_logo1.png")
    assert im.shape == (258, 540, 3)
    print(im.shape)

    im = imread("../tests/testdata/AE86.jpg")
    print(im.shape)
    assert im.shape == (193, 321, 3)

    pim = cv2pil(im)
    assert pim.size == (321, 193)

    taobao = imread("screenshot.jpg")
    import findit

    fi = findit.FindIt(engine=['template'],
                       engine_template_scale=(1, 1, 1),
                       pro_mode=True)
    fi.load_template("template", pic_object=taobao)

    import uiautomator2 as u2
    d = u2.connect()
    bg = d.screenshot(format="opencv")
    res = fi.find("target", target_pic_object=bg)
    from pprint import pprint
    pprint(res)
    # {'target_name': 'target',
    #  'target_path': None,
    #  'data': {
    #       'template': {
    #           'TemplateEngine': {
    #               'conf': {
    #                   'engine_template_cv_method_name': 'cv2.TM_CCOEFF_NORMED',
    #                   'engine_template_cv_method_code': 5,
    #                   'engine_template_scale': (1, 1, 1),
    #                   'engine_template_multi_target_max_threshold': 0.99,
    #                   'engine_template_multi_target_distance_threshold': 10.0,
    #                   'engine_template_compress_rate': 1.0},
    #               'target_point': [111, 1713],
    #               'target_sim': 0.9984192848205566,
    #               'raw': {'min_val': -0.4805332124233246,
    #               'max_val': 0.9984192848205566,
    #               'min_loc': [990, 1266],
    #               'max_loc': [111, 1713],
    #               'all': [[111.0, 1713.5]]},
    #               'ok': True}}}}
    x, y = res["data"]["template"]["TemplateEngine"]["target_point"]
    d.click(x, y)
